import os
from typing import Any, Optional

from testcontainers.core.waiting_utils import wait_for_logs

from tomodachi_testcontainers.containers.common import WebContainer
from tomodachi_testcontainers.utils import AWSClientConfig


class MotoContainer(WebContainer):
    def __init__(
        self,
        image: str = "motoserver/moto:latest",
        internal_port: int = 5000,
        edge_port: int = 5000,
        region_name: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(image, internal_port=internal_port, edge_port=edge_port, **kwargs)

        self.region_name = region_name or os.getenv("AWS_DEFAULT_REGION") or "us-east-1"
        self.aws_access_key_id = os.getenv("AWS_ACCESS_KEY_ID") or "testing"  # nosec: B105
        self.aws_secret_access_key = os.getenv("AWS_SECRET_ACCESS_KEY") or "testing"  # nosec: B105

        self.with_env("AWS_DEFAULT_REGION", self.region_name)
        self.with_env("AWS_ACCESS_KEY_ID", self.aws_access_key_id)
        self.with_env("AWS_SECRET_ACCESS_KEY", self.aws_secret_access_key)

        self.with_env("MOTO_PORT", str(self.internal_port))

        # Docker is needed for running AWS Lambda container
        self.with_env("MOTO_DOCKER_NETWORK_NAME", self.network)
        self.with_volume_mapping("/var/run/docker.sock", "/var/run/docker.sock")

    def log_message_on_container_start(self) -> str:
        return f"Moto dashboard: http://localhost:{self.edge_port}/moto-api"

    def get_aws_client_config(self) -> AWSClientConfig:
        return AWSClientConfig(
            region_name=self.region_name,
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
            endpoint_url=self.get_external_url(),
        )

    def start(self, timeout: float = 10.0, interval: float = 0.5, status_code: int = 200) -> "MotoContainer":
        super().start(timeout=timeout, interval=interval, status_code=status_code)
        wait_for_logs(self, "Running on all addresses", timeout=timeout)
        return self

    def reset_moto(self) -> None:
        self.exec(["curl", "-X", "POST", f"http://localhost:{self.internal_port}/moto-api/reset"])



import os
from typing import AsyncGenerator, Generator, cast

import pytest
import pytest_asyncio
from aiobotocore.session import get_session
from types_aiobotocore_dynamodb import DynamoDBClient
from types_aiobotocore_iam import IAMClient
from types_aiobotocore_lambda import LambdaClient
from types_aiobotocore_s3 import S3Client
from types_aiobotocore_sns import SNSClient
from types_aiobotocore_sqs import SQSClient
from types_aiobotocore_ssm import SSMClient

from tomodachi_testcontainers import LocalStackContainer
from tomodachi_testcontainers.clients.snssqs import SNSSQSTestClient
from tomodachi_testcontainers.utils import get_available_port


@pytest.fixture(scope="session")
def localstack_container() -> Generator[LocalStackContainer, None, None]:
    image = os.getenv("LOCALSTACK_TESTCONTAINER_IMAGE_ID", "localstack/localstack:3")
    with LocalStackContainer(image=image, edge_port=get_available_port()) as container:
        yield cast(LocalStackContainer, container)


@pytest.fixture()
def _restart_localstack_container_on_teardown(localstack_container: LocalStackContainer) -> Generator[None, None, None]:
    yield
    localstack_container.restart()


@pytest_asyncio.fixture(scope="session")
async def localstack_dynamodb_client(localstack_container: LocalStackContainer) -> AsyncGenerator[DynamoDBClient, None]:
    async with get_session().create_client("dynamodb", **localstack_container.get_aws_client_config()) as c:
        yield c


@pytest_asyncio.fixture(scope="session")
async def localstack_iam_client(localstack_container: LocalStackContainer) -> AsyncGenerator[IAMClient, None]:
    async with get_session().create_client("iam", **localstack_container.get_aws_client_config()) as c:
        yield c


@pytest_asyncio.fixture(scope="session")
async def localstack_lambda_client(localstack_container: LocalStackContainer) -> AsyncGenerator[LambdaClient, None]:
    async with get_session().create_client("lambda", **localstack_container.get_aws_client_config()) as c:
        yield c


@pytest_asyncio.fixture(scope="session")
async def localstack_s3_client(localstack_container: LocalStackContainer) -> AsyncGenerator[S3Client, None]:
    async with get_session().create_client("s3", **localstack_container.get_aws_client_config()) as c:
        yield c


@pytest_asyncio.fixture(scope="session")
async def localstack_sns_client(localstack_container: LocalStackContainer) -> AsyncGenerator[SNSClient, None]:
    async with get_session().create_client("sns", **localstack_container.get_aws_client_config()) as c:
        yield c


@pytest_asyncio.fixture(scope="session")
async def localstack_sqs_client(localstack_container: LocalStackContainer) -> AsyncGenerator[SQSClient, None]:
    async with get_session().create_client("sqs", **localstack_container.get_aws_client_config()) as c:
        yield c


@pytest_asyncio.fixture(scope="session")
async def localstack_ssm_client(localstack_container: LocalStackContainer) -> AsyncGenerator[SSMClient, None]:
    async with get_session().create_client("ssm", **localstack_container.get_aws_client_config()) as c:
        yield c


@pytest.fixture()
def localstack_snssqs_tc(localstack_sns_client: SNSClient, localstack_sqs_client: SQSClient) -> SNSSQSTestClient:
    return SNSSQSTestClient.create(localstack_sns_client, localstack_sqs_client)



import io
import logging
import os
import socket
import tarfile
from pathlib import Path
from typing import Dict, Optional, TypedDict, cast

from docker.errors import ImageNotFound
from docker.models.containers import Container
from docker.models.images import Image
from testcontainers.core.docker_client import DockerClient


class AWSClientConfig(TypedDict):
    region_name: str
    aws_access_key_id: str
    aws_secret_access_key: str
    endpoint_url: str


def get_docker_image(image_id: str, docker_client_kwargs: Optional[Dict] = None) -> Image:
    """Returns a Docker image, pulling it if not exists on host."""
    client = DockerClient(**(docker_client_kwargs or {}))
    try:
        return cast(Image, client.client.images.get(image_id))
    except ImageNotFound:
        return cast(Image, client.client.images.pull(image_id))


def copy_folder_to_container(container: Container, host_path: Path, container_path: Path) -> None:
    """Copies a folder from the host to the container."""
    tar_stream = io.BytesIO()
    with tarfile.open(fileobj=tar_stream, mode="w") as tar:
        for root, _, files in os.walk(host_path):
            for file in files:
                file_path = Path(root) / file
                arcname = os.path.relpath(file_path, host_path)
                tar.add(file_path, arcname=arcname)
    tar_stream.seek(0)
    container.put_archive(path=container_path, data=tar_stream)


def get_available_port() -> int:
    """Returns a random available port on the host."""
    with socket.socket() as sock:
        sock.bind(("", 0))
        return int(sock.getsockname()[1])


def get_current_ip_address() -> str:
    """Returns the current IP address of the host."""
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
        sock.connect(("8.8.8.8", 80))
        return str(sock.getsockname()[0])


def setup_logger(name: str) -> logging.Logger:
    """Outputs logs to stderr for better visibility in pytest output.

    Inspired by testcontainers.core.utils.setup_logger
    """
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler()
    handler.setLevel(logging.INFO)
    handler.setFormatter(logging.Formatter(fmt="%(name)s: %(message)s"))
    logger.addHandler(handler)
    return logger



"""MySQL testcontainer.

Adaptation of https://github.com/testcontainers/testcontainers-python/tree/main/mysql
"""
import os
from typing import Any, Optional

from tomodachi_testcontainers.containers.common import DatabaseContainer


class MySQLContainer(DatabaseContainer):
    def __init__(
        self,
        image: str = "mysql:8",
        internal_port: int = 3306,
        edge_port: int = 3306,
        drivername: Optional[str] = None,
        username: Optional[str] = None,
        root_password: Optional[str] = None,
        password: Optional[str] = None,
        database: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(image, internal_port=internal_port, edge_port=edge_port, **kwargs)

        self.drivername = drivername or os.getenv("MYSQL_DRIVERNAME") or "mysql+pymysql"
        self.username = username or os.getenv("MYSQL_USER") or "username"
        self.root_password = root_password or os.getenv("MYSQL_ROOT_PASSWORD") or "root"
        self.password = password or os.getenv("MYSQL_PASSWORD") or "password"
        self.database = database or os.getenv("MYSQL_DATABASE") or "db"

        if self.username == "root":
            self.root_password = self.password

        self.with_env("MYSQL_USER", self.username)
        self.with_env("MYSQL_ROOT_PASSWORD", self.root_password)
        self.with_env("MYSQL_PASSWORD", self.password)
        self.with_env("MYSQL_DATABASE", self.database)

        # Do not flush data on disk to improve test container performance
        # https://pythonspeed.com/articles/faster-db-tests/
        self.with_command("--innodb_flush_method=O_DIRECT_NO_FSYNC")

    def log_message_on_container_start(self) -> str:
        return f"MySQL started: {self.get_external_url()}"



"""Abstract web container for services that expose HTTP port."""
import abc
import urllib.parse
from typing import Any, Optional

import requests
from tenacity import Retrying
from tenacity.stop import stop_after_delay
from tenacity.wait import wait_fixed

from tomodachi_testcontainers.containers.common.container import DockerContainer


class WebContainer(DockerContainer, abc.ABC):
    internal_port: int
    edge_port: int

    def __init__(
        self,
        image: str,
        internal_port: int,
        edge_port: int,
        http_healthcheck_path: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(image, **kwargs)
        self.internal_port = internal_port
        self.edge_port = edge_port
        self.http_healthcheck_path = http_healthcheck_path
        self.with_bind_ports(internal_port, edge_port)

    def get_internal_url(self) -> str:
        ip = self.get_container_internal_ip()
        return f"http://{ip}:{self.internal_port}"

    def get_external_url(self) -> str:
        host = self.get_container_host_ip()
        return f"http://{host}:{self.edge_port}"

    def start(self, timeout: float = 10.0, interval: float = 0.5, status_code: int = 200) -> "WebContainer":
        super().start()
        if self.http_healthcheck_path:
            url = urllib.parse.urljoin(self.get_external_url(), self.http_healthcheck_path)
            wait_for_http_healthcheck(url=url, timeout=timeout, interval=interval, status_code=status_code)
        return self


def wait_for_http_healthcheck(url: str, timeout: float = 10.0, interval: float = 0.5, status_code: int = 200) -> None:
    for attempt in Retrying(stop=stop_after_delay(timeout), wait=wait_fixed(interval), reraise=True):
        with attempt:
            response = requests.get(url, timeout=timeout)
            if response.status_code != status_code:
                raise RuntimeError(f"Healthcheck failed with HTTP status code: {response.status_code}")



import abc
import contextlib
import logging
import os
from types import TracebackType
from typing import Any, Dict, Optional, Type, cast

import docker.errors
import shortuuid
import testcontainers.core.container
from docker.models.containers import Container
from testcontainers.core.utils import inside_container

from tomodachi_testcontainers.utils import setup_logger


class ContainerWithSameNameAlreadyExistsError(Exception):
    pass


class DockerContainer(testcontainers.core.container.DockerContainer, abc.ABC):
    _container: Optional[Container]
    _name: str
    _logger: logging.Logger

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.network = os.getenv("TESTCONTAINER_DOCKER_NETWORK") or "bridge"
        super().__init__(*args, **kwargs, network=self.network)

    def __enter__(self) -> "DockerContainer":
        try:
            return self.start()
        except ContainerWithSameNameAlreadyExistsError:
            raise
        except Exception:
            self._forward_container_logs_to_logger()
            self.stop()
            raise

    def __exit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> None:
        self._forward_container_logs_to_logger()
        self.stop()

    @abc.abstractmethod
    def log_message_on_container_start(self) -> str:
        """Returns a message that will be logged when the container starts."""

    def get_container_host_ip(self) -> str:
        host = self.get_docker_client().host()
        if not host:
            return "localhost"
        if inside_container() and not os.getenv("DOCKER_HOST"):
            gateway_ip = self.get_container_gateway_ip()
            if gateway_ip == host:
                return self.get_container_internal_ip()
            return gateway_ip
        return host

    def get_container_internal_ip(self) -> str:
        return self.docker_inspect()["NetworkSettings"]["Networks"][self.network]["IPAddress"]

    def get_container_gateway_ip(self) -> str:
        return self.docker_inspect()["NetworkSettings"]["Networks"][self.network]["Gateway"]

    def docker_inspect(self) -> Dict[str, Any]:
        return self.get_docker_client().get_container(self.get_wrapped_container().id)

    def start(self) -> "DockerContainer":
        self._set_container_name()
        self._setup_logger()
        self._start()
        self._log_message_on_container_start()
        return self

    def stop(self) -> None:
        with contextlib.suppress(Exception):
            container = self._container or cast(Container, self.get_docker_client().client.containers.get(self._name))
            container.remove(force=True, v=True)
        self._container = None

    def restart(self) -> None:
        self.get_wrapped_container().restart()

    def _set_container_name(self) -> None:
        self._name = self._name or f"testcontainer-{shortuuid.uuid()}"

    def _setup_logger(self) -> None:
        self._logger = setup_logger(f"{self.__class__.__name__} ({self._name})")

    def _start(self) -> None:
        self._logger.info(f"Pulling image: {self.image}")
        try:
            self._container = self.get_docker_client().run(
                image=self.image,
                command=self._command or "",
                detach=True,
                environment=self.env,
                ports=self.ports,
                name=self._name,
                volumes=self.volumes,
                **self._kwargs,
            )
        except Exception as exc:
            self._logger.exception("Failed to start the container")
            if isinstance(exc, docker.errors.APIError) and exc.status_code == 409:  # pylint: disable=no-member
                raise ContainerWithSameNameAlreadyExistsError(self._name) from exc
            raise
        else:
            self._logger.info(f"Container started: {self._name}")

    def _log_message_on_container_start(self) -> None:
        if message := self.log_message_on_container_start():
            self._logger.info(message)

    def _forward_container_logs_to_logger(self) -> None:
        if container := self.get_wrapped_container():
            logs = bytes(container.logs(timestamps=True)).decode().split("\n")
            for log in logs:
                self._logger.info(log)



import contextlib

from tomodachi_testcontainers.containers.common import DockerContainer, EphemeralDockerImage, WebContainer
from tomodachi_testcontainers.containers.localstack import LocalStackContainer
from tomodachi_testcontainers.containers.moto import MotoContainer
from tomodachi_testcontainers.containers.tomodachi import TomodachiContainer
from tomodachi_testcontainers.containers.wiremock import WireMockContainer

with contextlib.suppress(ImportError):  # 'db' extra dependency
    from tomodachi_testcontainers.containers.common import DatabaseContainer

with contextlib.suppress(ImportError):  # 'mysql' extra dependency
    from tomodachi_testcontainers.containers.mysql import MySQLContainer

with contextlib.suppress(ImportError):  # 'postgres' extra dependency
    from tomodachi_testcontainers.containers.postgres import PostgreSQLContainer

with contextlib.suppress(ImportError):  # 'sftp' extra dependency
    from tomodachi_testcontainers.containers.sftp import SFTPContainer

__all__ = [
    "DatabaseContainer",
    "DockerContainer",
    "EphemeralDockerImage",
    "LocalStackContainer",
    "MotoContainer",
    "MySQLContainer",
    "PostgreSQLContainer",
    "SFTPContainer",
    "TomodachiContainer",
    "WebContainer",
    "WireMockContainer",
]



from datetime import datetime, timedelta, timezone
from typing import Tuple, cast

from tomodachi_testcontainers import DockerContainer

UUID4_PATTERN = r"[0-9a-f]{8}-?[0-9a-f]{4}-?4[0-9a-f]{3}-?[89ab][0-9a-f]{3}-?[0-9a-f]{12}"

DEFAULT_DATETIME_RANGE = timedelta(seconds=10)


def assert_datetime_within_range(value: datetime, range: timedelta = DEFAULT_DATETIME_RANGE) -> None:
    current_datetime = datetime.utcnow().replace(tzinfo=timezone.utc)
    start_datetime = current_datetime - range
    end_datetime = current_datetime + range
    assert start_datetime <= value <= end_datetime  # nosec: B101


def assert_logs_contain(container: DockerContainer, contains: str) -> None:
    stdout_and_stderr_logs = cast(Tuple[bytes, bytes], container.get_logs())
    for log in stdout_and_stderr_logs:
        if contains in log.decode():
            return
    raise AssertionError(f"Expected logs to contain: '{contains}'")


def assert_logs_not_contain(container: DockerContainer, contains: str) -> None:
    stdout_and_stderr_logs = cast(Tuple[bytes, bytes], container.get_logs())
    for log in stdout_and_stderr_logs:
        if contains in log.decode():
            raise AssertionError(f"Expected logs not to contain: '{contains}'")


def assert_logs_match_line_count(container: DockerContainer, contains: str, count: int) -> None:
    stdout_logs = cast(bytes, container.get_logs()[0])
    stderr_logs = cast(bytes, container.get_logs()[1])
    logs = "\n".join([stdout_logs.decode(), stderr_logs.decode()])
    matched_lines = [log for log in logs.splitlines() if contains in log]
    error_msg = f"Expected '{contains}' to be contained in {count} lines, found {len(matched_lines)} lines"
    assert len(matched_lines) == count, error_msg  # nosec: B101



import os
from typing import Generator, cast

import pytest

from tomodachi_testcontainers import MySQLContainer
from tomodachi_testcontainers.utils import get_available_port


@pytest.fixture(scope="session")
def mysql_container() -> Generator[MySQLContainer, None, None]:
    image = os.getenv("MYSQL_TESTCONTAINER_IMAGE_ID", "mysql:8")
    with MySQLContainer(image=image, edge_port=get_available_port()) as container:
        yield cast(MySQLContainer, container)



import os
import subprocess  # nosec: B404
from pathlib import Path
from types import TracebackType
from typing import Dict, Iterator, Optional, Tuple, Type, cast

from docker.models.images import Image
from testcontainers.core.docker_client import DockerClient


class EphemeralDockerImage:
    """Builds a Docker image from a given Dockerfile and removes it when the context manager exits."""

    image: Image

    def __init__(
        self,
        dockerfile: Optional[Path] = None,
        context: Optional[Path] = None,
        target: Optional[str] = None,
        docker_client_kwargs: Optional[Dict] = None,
    ) -> None:
        self.dockerfile = str(dockerfile) if dockerfile else None
        self.context = str(context) if context else "."
        self.target = target
        self._docker_client = DockerClient(**(docker_client_kwargs or {}))

    def __enter__(self) -> Image:
        self._build_image()
        return self.image

    def __exit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> None:
        self._remove_image()

    def _build_image(self) -> None:
        if os.getenv("DOCKER_BUILDKIT"):
            self.image = self._build_with_docker_buildkit()
        else:
            self.image = self._build_with_docker_client()

    def _remove_image(self) -> None:
        self._docker_client.client.images.remove(image=str(self.image.id))

    def _build_with_docker_buildkit(self) -> Image:
        cmd = ["docker", "build", "-q", "--rm=true"]
        if self.dockerfile:
            cmd.extend(["-f", self.dockerfile])
        if self.target:
            cmd.extend(["--target", self.target])
        cmd.append(self.context)

        result = subprocess.run(  # nosec: B603
            cmd,
            shell=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=True,
        )
        image_id = result.stdout.decode("utf-8").strip()
        return cast(Image, self._docker_client.client.images.get(image_id))

    def _build_with_docker_client(self) -> Image:
        image, _ = cast(
            Tuple[Image, Iterator],
            self._docker_client.client.images.build(
                dockerfile=self.dockerfile,
                path=self.context,
                target=self.target,
                forcerm=True,
            ),
        )
        return image



from typing import Any, Optional

from testcontainers.core.waiting_utils import wait_for_logs

from tomodachi_testcontainers.containers.common import WebContainer


class TomodachiContainer(WebContainer):
    def __init__(
        self,
        image: str,
        internal_port: int = 9700,
        edge_port: int = 9700,
        http_healthcheck_path: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            image=image,
            internal_port=internal_port,
            edge_port=edge_port,
            http_healthcheck_path=http_healthcheck_path,
            **kwargs,
        )

    def log_message_on_container_start(self) -> str:
        return f"Tomodachi service: http://localhost:{self.edge_port}"

    def start(self, timeout: float = 10.0, interval: float = 0.5, status_code: int = 200) -> "TomodachiContainer":
        super().start(timeout=timeout, interval=interval, status_code=status_code)
        # Apart from HTTP healthcheck, we need to wait for "started service" log
        # to make sure messaging transport like AWS SNS SQS is also up and running.
        # It's started independently from HTTP transport.

        # Different service start messages depending on tomodachi version
        # tomodachi < 0.26.0 - "Started service"
        # tomodachi >= 0.26.0 - "started service successfully"
        # using (?i) to ignore case to support both versions
        wait_for_logs(self, "(?i)started service", timeout=timeout)
        return self



import os
from typing import Generator, cast

import pytest

from tomodachi_testcontainers import PostgreSQLContainer
from tomodachi_testcontainers.utils import get_available_port


@pytest.fixture(scope="session")
def postgres_container() -> Generator[PostgreSQLContainer, None, None]:
    image = os.getenv("POSTGRES_TESTCONTAINER_IMAGE_ID", "postgres:16")
    with PostgreSQLContainer(image=image, edge_port=get_available_port()) as container:
        yield cast(PostgreSQLContainer, container)



"""PostgreSQL testcontainer.

Adaptation of https://github.com/testcontainers/testcontainers-python/tree/main/postgres
"""
import os
from typing import Any, Optional

from tomodachi_testcontainers.containers.common import DatabaseContainer


class PostgreSQLContainer(DatabaseContainer):
    def __init__(
        self,
        image: str = "postgres:16",
        internal_port: int = 5432,
        edge_port: int = 5432,
        drivername: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        database: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(image, internal_port=internal_port, edge_port=edge_port, **kwargs)

        self.drivername = drivername or os.getenv("POSTGRES_DRIVERNAME") or "postgresql+psycopg2"
        self.username = username or os.getenv("POSTGRES_USER") or "username"
        self.password = password or os.getenv("POSTGRES_PASSWORD") or "password"
        self.database = database or os.getenv("POSTGRES_DB") or "db"

        self.with_env("POSTGRES_USER", self.username)
        self.with_env("POSTGRES_PASSWORD", self.password)
        self.with_env("POSTGRES_DB", self.database)

        # Do not flush data on disk to improve test container performance
        # https://pythonspeed.com/articles/faster-db-tests/
        self.with_command("-c fsync=off")

    def log_message_on_container_start(self) -> str:
        return f"PostgreSQL started: {self.get_external_url()}"



import inspect
import json
from contextlib import suppress
from typing import Any, Dict, List, Optional, Protocol, Type, TypeVar, Union

from botocore.exceptions import ClientError
from google.protobuf.message import Message
from types_aiobotocore_sns import SNSClient
from types_aiobotocore_sns.type_defs import MessageAttributeValueTypeDef
from types_aiobotocore_sqs import SQSClient
from types_aiobotocore_sqs.literals import QueueAttributeFilterType, QueueAttributeNameType

__all__ = [
    "SNSSQSTestClient",
]

MessageType = TypeVar("MessageType")

TopicARNType = str

QueueARNType = str
QueueURLType = str


class TopicDoesNotExist(Exception):
    pass  # pragma: no cover


class QueueDoesNotExist(Exception):
    pass  # pragma: no cover


class _TomodachiSNSSQSEnvelopeStatic(Protocol):
    @classmethod
    async def build_message(
        cls: "_TomodachiSNSSQSEnvelopeStatic", service: Any, topic: str, data: Any, **kwargs: Any
    ) -> str:
        ...  # pragma: no cover

    @classmethod
    async def parse_message(cls: "_TomodachiSNSSQSEnvelopeStatic", payload: str, **kwargs: Any) -> Union[dict, tuple]:
        ...  # pragma: no cover


class _TomodachiSNSSQSEnvelopeInstance(Protocol):
    async def build_message(
        self: "_TomodachiSNSSQSEnvelopeInstance", service: Any, topic: str, data: Any, **kwargs: Any
    ) -> str:
        ...  # pragma: no cover

    async def parse_message(
        self: "_TomodachiSNSSQSEnvelopeInstance", payload: str, **kwargs: Any
    ) -> Union[dict, tuple]:
        ...  # pragma: no cover


TomodachiSNSSQSEnvelope = Union[_TomodachiSNSSQSEnvelopeStatic, _TomodachiSNSSQSEnvelopeInstance]


class SNSSQSTestClientCache:
    def __init__(self) -> None:
        self.hit_count = 0
        self._topic_arn: Dict[str, TopicARNType] = {}
        self._queue_arn: Dict[str, QueueARNType] = {}
        self._queue_url: Dict[str, QueueURLType] = {}

    def clear(self) -> None:
        self.hit_count = 0
        self._topic_arn.clear()
        self._queue_arn.clear()
        self._queue_url.clear()

    def save_topic_arn(self, topic: str, arn: TopicARNType) -> None:
        self._topic_arn[topic] = arn

    def save_queue_arn(self, queue: str, arn: QueueARNType) -> None:
        self._queue_arn[queue] = arn

    def save_queue_url(self, queue: str, url: QueueURLType) -> None:
        self._queue_url[queue] = url

    def get_topic_arn(self, topic: str) -> Optional[str]:
        if arn := self._topic_arn.get(topic):
            self.hit_count += 1
        return arn

    def get_queue_arn(self, queue: str) -> Optional[str]:
        if arn := self._queue_arn.get(queue):
            self.hit_count += 1
        return arn

    def get_queue_url(self, queue: str) -> Optional[str]:
        if url := self._queue_url.get(queue):
            self.hit_count += 1
        return url


class SNSSQSTestClient:
    """Wraps aiobotocore SNS and SQS clients and provides common methods for testing SNS SQS integrations."""

    def __init__(self, sns_client: SNSClient, sqs_client: SQSClient, cache: SNSSQSTestClientCache) -> None:
        self._sns_client = sns_client
        self._sqs_client = sqs_client
        self._cache = cache

    @staticmethod
    def create(sns_client: SNSClient, sqs_client: SQSClient) -> "SNSSQSTestClient":
        return SNSSQSTestClient(sns_client, sqs_client, SNSSQSTestClientCache())

    async def create_topic(self, topic: str, *, fifo: bool = False) -> TopicARNType:
        with suppress(TopicDoesNotExist):
            return await self.get_topic_arn(topic)
        topic_attributes: Dict[str, str] = {}
        if fifo:
            topic_attributes.update(
                {
                    "FifoTopic": "true",
                    "ContentBasedDeduplication": "false",
                }
            )
        create_topic_response = await self._sns_client.create_topic(Name=topic, Attributes=topic_attributes)
        topic_arn = create_topic_response["TopicArn"]
        self._cache.save_topic_arn(topic, topic_arn)
        return topic_arn

    async def create_queue(self, queue: str, *, fifo: bool = False) -> QueueARNType:
        with suppress(QueueDoesNotExist):
            return await self.get_queue_arn(queue)
        queue_attributes: Dict[QueueAttributeNameType, str] = {}
        if fifo:
            queue_attributes.update(
                {
                    "FifoQueue": "true",
                    "ContentBasedDeduplication": "false",
                }
            )
        await self._sqs_client.create_queue(QueueName=queue, Attributes=queue_attributes)
        queue_attributes = await self.get_queue_attributes(queue, attributes=["QueueArn"])
        queue_arn = queue_attributes["QueueArn"]
        self._cache.save_queue_arn(queue, queue_arn)
        return queue_arn

    async def subscribe_to(
        self,
        topic: str,
        queue: str,
        subscribe_attributes: Optional[Dict[str, str]] = None,
        *,
        fifo: bool = False,
    ) -> None:
        """Subscribe a SQS queue to a SNS topic; create the topic and queue if they don't exist."""
        topic_arn = await self.create_topic(topic, fifo=fifo)
        queue_arn = await self.create_queue(queue, fifo=fifo)
        await self._sns_client.subscribe(
            TopicArn=topic_arn,
            Protocol="sqs",
            Endpoint=queue_arn,
            Attributes=subscribe_attributes or {},
        )

    async def receive(
        self, queue: str, envelope: TomodachiSNSSQSEnvelope, message_type: Type[MessageType], max_messages: int = 10
    ) -> List[MessageType]:
        queue_url = await self.get_queue_url(queue)
        received_messages_response = await self._sqs_client.receive_message(
            QueueUrl=queue_url, MaxNumberOfMessages=max_messages
        )
        received_messages = received_messages_response.get("Messages")
        if not received_messages:
            return []

        if inspect.isclass(message_type) and issubclass(message_type, Message):
            proto_class = message_type
        else:
            proto_class = None

        parsed_messages: List[MessageType] = []
        for received_message in received_messages:
            payload = json.loads(received_message["Body"])["Message"]
            parsed_message = await envelope.parse_message(payload=payload, proto_class=proto_class)
            parsed_messages.append(parsed_message[0]["data"])
            await self._sqs_client.delete_message(QueueUrl=queue_url, ReceiptHandle=received_message["ReceiptHandle"])
        return parsed_messages

    async def publish(
        self,
        topic: str,
        data: Any,
        envelope: TomodachiSNSSQSEnvelope,
        message_attributes: Optional[Dict[str, MessageAttributeValueTypeDef]] = None,
        message_deduplication_id: Optional[str] = None,
        message_group_id: Optional[str] = None,
    ) -> None:
        topic_arn = await self.get_topic_arn(topic)
        message = await envelope.build_message(service={}, topic=topic, data=data)
        sns_publish_kwargs: Dict[str, Any] = {}
        if message_attributes:
            sns_publish_kwargs["MessageAttributes"] = message_attributes
        if message_deduplication_id:
            sns_publish_kwargs["MessageDeduplicationId"] = message_deduplication_id
        if message_group_id:
            sns_publish_kwargs["MessageGroupId"] = message_group_id
        await self._sns_client.publish(TopicArn=topic_arn, Message=message, **sns_publish_kwargs)

    async def get_topic_arn(self, topic: str) -> str:
        if topic_arn := self._cache.get_topic_arn(topic):
            return topic_arn
        list_topics_response = await self._sns_client.list_topics()
        topic_arn = next((v["TopicArn"] for v in list_topics_response["Topics"] if v["TopicArn"].endswith(topic)), None)
        if not topic_arn:
            raise TopicDoesNotExist(topic)
        self._cache.save_topic_arn(topic, topic_arn)
        return topic_arn

    async def get_topic_attributes(self, topic: str) -> Dict[str, str]:
        topic_arn = await self.get_topic_arn(topic)
        get_topic_attributes_response = await self._sns_client.get_topic_attributes(TopicArn=topic_arn)
        return get_topic_attributes_response["Attributes"]

    async def get_queue_arn(self, queue: str) -> str:
        if queue_arn := self._cache.get_queue_arn(queue):
            return queue_arn
        attributes = await self.get_queue_attributes(queue, attributes=["QueueArn"])
        queue_arn = attributes["QueueArn"]
        self._cache.save_queue_arn(queue, queue_arn)
        return queue_arn

    async def get_queue_url(self, queue: str) -> str:
        if queue_url := self._cache.get_queue_url(queue):
            return queue_url
        try:
            get_queue_response = await self._sqs_client.get_queue_url(QueueName=queue)
            queue_url = get_queue_response["QueueUrl"]
            self._cache.save_queue_url(queue, queue_url)
            return queue_url
        except ClientError as exc:
            raise QueueDoesNotExist(queue) from exc

    async def get_queue_attributes(
        self, queue: str, attributes: List[QueueAttributeFilterType]
    ) -> Dict[QueueAttributeNameType, str]:
        queue_url = await self.get_queue_url(queue)
        get_queue_attributes_response = await self._sqs_client.get_queue_attributes(
            QueueUrl=queue_url, AttributeNames=attributes
        )
        return get_queue_attributes_response["Attributes"]

    async def purge_queue(self, queue: str) -> None:
        queue_url = await self.get_queue_url(queue)
        await self._sqs_client.purge_queue(QueueUrl=queue_url)



import contextlib

from tomodachi_testcontainers.containers.common.container import DockerContainer
from tomodachi_testcontainers.containers.common.image import EphemeralDockerImage
from tomodachi_testcontainers.containers.common.web import WebContainer

with contextlib.suppress(ImportError):  # 'db' extra dependency
    from tomodachi_testcontainers.containers.common.database import DatabaseContainer

__all__ = [
    "DatabaseContainer",
    "DockerContainer",
    "EphemeralDockerImage",
    "WebContainer",
]



"""Abstract relational database container."""

import abc
from typing import Any, NamedTuple

import sqlalchemy
from tenacity import Retrying
from tenacity.stop import stop_after_delay
from tenacity.wait import wait_fixed

from tomodachi_testcontainers.containers.common.container import DockerContainer


class DatabaseURL(NamedTuple):
    drivername: str
    username: str
    password: str
    host: str
    port: int
    database: str

    def __repr__(self) -> str:
        return self.to_str()

    def to_str(self) -> str:
        return f"{self.drivername}://{self.username}:{self.password}@{self.host}:{self.port}/{self.database}"


class DatabaseContainer(DockerContainer, abc.ABC):
    internal_port: int
    edge_port: int

    drivername: str
    username: str
    password: str
    database: str

    def __init__(self, image: str, internal_port: int, edge_port: int, **kwargs: Any) -> None:
        super().__init__(image, **kwargs)
        self.internal_port = internal_port
        self.edge_port = edge_port
        self.with_bind_ports(internal_port, edge_port)

    def get_internal_url(self) -> DatabaseURL:
        return DatabaseURL(
            drivername=self.drivername,
            username=self.username,
            password=self.password,
            host=self.get_container_internal_ip(),
            port=self.internal_port,
            database=self.database,
        )

    def get_external_url(self) -> DatabaseURL:
        return DatabaseURL(
            drivername=self.drivername,
            username=self.username,
            password=self.password,
            host=self.get_container_host_ip(),
            port=self.edge_port,
            database=self.database,
        )

    def start(self, timeout: float = 20.0) -> "DatabaseContainer":
        super().start()
        wait_for_database_healthcheck(url=self.get_external_url(), timeout=timeout)
        return self


def wait_for_database_healthcheck(url: DatabaseURL, timeout: float = 20.0, interval: float = 0.5) -> None:
    for attempt in Retrying(stop=stop_after_delay(timeout), wait=wait_fixed(interval), reraise=True):
        with attempt:
            engine = sqlalchemy.create_engine(url.to_str())
            with engine.connect() as conn:
                conn.execute(sqlalchemy.text("SELECT 1;"))



"""WireMock testcontainer.

Adaptation of https://github.com/wiremock/python-wiremock
See WireMock usage examples in https://github.com/wiremock/python-wiremock/tree/master/examples
"""
from pathlib import Path
from typing import Any, Optional

from testcontainers.core.waiting_utils import wait_for_logs

from tomodachi_testcontainers.containers.common import WebContainer
from tomodachi_testcontainers.utils import copy_folder_to_container


class WireMockContainer(WebContainer):
    MAPPINGS_DIR: Path = Path("/home/wiremock/mappings/")
    FILES_DIR: Path = Path("/home/wiremock/__files/")

    def __init__(
        self,
        mapping_stubs: Optional[Path] = None,
        mapping_files: Optional[Path] = None,
        image: str = "wiremock/wiremock:latest",
        internal_port: int = 8080,
        edge_port: int = 8080,
        *,
        verbose: bool = False,
        **kwargs: Any,
    ) -> None:
        super().__init__(image, internal_port=internal_port, edge_port=edge_port, **kwargs)

        self.mapping_stubs = mapping_stubs
        self.mapping_files = mapping_files

        if verbose:
            self.with_command("--verbose")

    def log_message_on_container_start(self) -> str:
        return f"Wiremock admin: http://localhost:{self.edge_port}/__admin"

    def start(self, timeout: float = 10.0, interval: float = 0.5, status_code: int = 200) -> "WireMockContainer":
        super().start(timeout=timeout, interval=interval, status_code=status_code)
        wait_for_logs(self, "port:", timeout=timeout)
        self.copy_mappings_to_container()
        self.copy_mapping_files_to_container()
        self.reload_mappings()
        return self

    def copy_mappings_to_container(self) -> None:
        if self.mapping_stubs is not None:
            copy_folder_to_container(
                self.get_wrapped_container(), host_path=self.mapping_stubs, container_path=self.MAPPINGS_DIR
            )

    def copy_mapping_files_to_container(self) -> None:
        if self.mapping_files is not None:
            copy_folder_to_container(
                self.get_wrapped_container(), host_path=self.mapping_files, container_path=self.FILES_DIR
            )

    def reload_mappings(self) -> None:
        self.exec(["curl", "-X", "POST", f"http://localhost:{self.internal_port}/__admin/mappings/reset"])
